import pytest

from bumpversion.functions import NumericFunction, ValuesFunction

# NumericFunction


def test_numeric_init_wo_first_value():
    func = NumericFunction()
    assert func.first_value == "0"


def test_numeric_init_w_first_value():
    func = NumericFunction(first_value="5")
    assert func.first_value == "5"


def test_numeric_init_non_numeric_first_value():
    with pytest.raises(ValueError):
        NumericFunction(first_value="a")


def test_numeric_bump_simple_number():
    func = NumericFunction()
    assert func.bump("0") == "1"


def test_numeric_bump_prefix_and_suffix():
    func = NumericFunction()
    assert func.bump("v0b") == "v1b"


# ValuesFunction


def test_values_init():
    func = ValuesFunction(["0", "1", "2"])
    assert func.optional_value == "0"
    assert func.first_value == "0"


def test_values_init_w_correct_optional_value():
    func = ValuesFunction(["0", "1", "2"], optional_value="1")
    assert func.optional_value == "1"
    assert func.first_value == "0"


def test_values_init_w_correct_first_value():
    func = ValuesFunction(["0", "1", "2"], first_value="1")
    assert func.optional_value == "0"
    assert func.first_value == "1"


def test_values_init_w_correct_optional_and_first_value():
    func = ValuesFunction(["0", "1", "2"], optional_value="0", first_value="1")
    assert func.optional_value == "0"
    assert func.first_value == "1"


def test_values_init_w_empty_values():
    with pytest.raises(ValueError):
        ValuesFunction([])


def test_values_init_w_incorrect_optional_value():
    with pytest.raises(ValueError):
        ValuesFunction(["0", "1", "2"], optional_value="3")


def test_values_init_w_incorrect_first_value():
    with pytest.raises(ValueError):
        ValuesFunction(["0", "1", "2"], first_value="3")


def test_values_bump():
    func = ValuesFunction(["0", "5", "10"])
    assert func.bump("0") == "5"


def test_values_bump_max():
    func = ValuesFunction(["0", "5", "10"])
    with pytest.raises(ValueError):
        func.bump("10")



"""File processing tests."""
import os
from datetime import datetime, timezone
from pathlib import Path
from textwrap import dedent

import pytest
from pytest import param

from bumpversion import exceptions, files
from bumpversion.utils import get_context
from tests.conftest import get_config_data, inside_dir


@pytest.mark.parametrize(
    ["glob_pattern", "file_list"],
    [
        param("*.txt", {Path("file1.txt"), Path("file2.txt")}, id="simple-glob"),
        param("**/*.txt", {Path("file1.txt"), Path("file2.txt"), Path("directory/file3.txt")}, id="recursive-glob"),
    ],
)
def test_get_glob_files(glob_pattern: str, file_list: set, fixtures_path: Path):
    """Get glob files should return all the globbed files and nothing else."""
    overrides = {
        "current_version": "1.0.0",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<release>[a-z]+))?",
        "serialize": ["{major}.{minor}.{release}", "{major}.{minor}"],
        "files": [
            {
                "glob": glob_pattern,
            }
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    with inside_dir(fixtures_path.joinpath("glob")):
        result = files.get_glob_files(conf.files[0], version_config)

    assert len(result) == len(file_list)
    for f in result:
        assert Path(f.path) in file_list


def test_single_file_processed_twice(tmp_path: Path):
    """
    Verify that a single file "file2" can be processed twice.

    Use two file_ entries, both with a different suffix after
    the underscore.
    Employ different parse/serialize and search/replace configs
    to verify correct interpretation.
    """
    filepath = tmp_path.joinpath("file2")
    filepath.write_text("dots: 0.10.2\ndashes: 0-10-2")
    overrides = {
        "current_version": "0.10.2",
        "files": [
            {
                "filename": str(filepath),
                "search": "dots: {current_version}",
                "replace": "dots: {new_version}",
            },
            {
                "filename": str(filepath),
                "search": "dashes: {current_version}",
                "replace": "dashes: {new_version}",
                "parse": r"(?P<major>\d+)-(?P<minor>\d+)-(?P<patch>\d+)",
                "serialize": ["{major}-{minor}-{patch}"],
            },
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch", version_config.order)
    ctx = get_context(conf)

    assert len(conf.files) == 2
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.replace_version(current_version, new_version, ctx)

    assert filepath.read_text() == "dots: 0.10.3\ndashes: 0-10-3"


def test_multi_file_configuration(tmp_path: Path):
    full_vers_path = tmp_path / "FULL_VERSION.txt"
    full_vers_path.write_text("1.0.3")
    maj_vers_path = tmp_path / "MAJOR_VERSION.txt"
    maj_vers_path.write_text("1")
    readme_path = tmp_path / "README.txt"
    readme_path.write_text("MyAwesomeSoftware(TM) v1.0")
    build_num_path = tmp_path / "BUILD_NUMBER"
    build_num_path.write_text("1.0.3+joe+38943")

    overrides = {
        "current_version": "1.0.3+joe+38943",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:\+.*)?",
        "serialize": [
            "{major}.{minor}.{patch}+{$USER}+{$BUILD_NUMBER}",
            "{major}.{minor}.{patch}",
        ],
        "files": [
            {
                "filename": str(full_vers_path),
            },
            {
                "filename": str(maj_vers_path),
                "parse": r"\d+",
                "serialize": ["{major}"],
            },
            {
                "filename": str(readme_path),
                "parse": r"(?P<major>\d+)\.(?P<minor>\d+)",
                "serialize": ["{major}.{minor}"],
            },
            {
                "filename": str(build_num_path),
                "serialize": ["{major}.{minor}.{patch}+{$USER}+{$BUILD_NUMBER}"],
            },
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    major_version = current_version.bump("major", version_config.order)

    os.environ["BUILD_NUMBER"] = "38944"
    os.environ["USER"] = "bob"
    ctx = get_context(conf)
    del os.environ["BUILD_NUMBER"]
    del os.environ["USER"]

    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.replace_version(current_version, major_version, ctx)

    assert full_vers_path.read_text() == "2.0.0"
    assert maj_vers_path.read_text() == "2"
    assert readme_path.read_text() == "MyAwesomeSoftware(TM) v2.0"
    assert build_num_path.read_text() == "2.0.0+bob+38944"

    os.environ["BUILD_NUMBER"] = "38945"
    os.environ["USER"] = "jane"
    conf.current_version = "2.0.0+bob+38944"
    ctx = get_context(conf)
    del os.environ["BUILD_NUMBER"]
    del os.environ["USER"]
    major_version = version_config.parse(conf.current_version)
    major_patch_version = major_version.bump("patch", version_config.order)
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.replace_version(major_version, major_patch_version, ctx)

    assert full_vers_path.read_text() == "2.0.1"
    assert maj_vers_path.read_text() == "2"
    assert readme_path.read_text() == "MyAwesomeSoftware(TM) v2.0"
    assert build_num_path.read_text() == "2.0.1+jane+38945"


def test_issue_14(tmp_path: Path):
    full_vers_path = tmp_path / "FULL_VERSION.txt"
    full_vers_path.write_text("3.1.0-rc+build.1031")
    assembly_path = tmp_path / "AssemblyInfo.cs"
    assembly_path.write_text(
        '[assembly: AssemblyFileVersion("3.1.0-rc+build.1031")]\n' '[assembly: AssemblyVersion("3.1.1031.0")]'
    )
    csv_path = tmp_path / "Version.csv"
    csv_path.write_text("1;3;1;0;rc;build.1031")

    overrides = {
        "current_version": "3.1.0-rc+build.1031",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(-(?P<release>[0-9A-Za-z]+))?(\+build\.(?P<build>.[0-9A-Za-z]+))?",
        "serialize": ["{major}.{minor}.{patch}-{release}+build.{build}", "{major}.{minor}.{patch}+build.{build}"],
        "commit": True,
        "message": "Bump version: {current_version} -> {new_version}",
        "tag": False,
        "tag_name": "{new_version}",
        "tag_message": "Version {new_version}",
        "allow_dirty": True,
        "files": [
            {
                "filename": str(full_vers_path),
            },
            {
                "filename": str(assembly_path),
                "search": '[assembly: AssemblyFileVersion("{current_version}")]',
                "replace": '[assembly: AssemblyFileVersion("{new_version}")]',
            },
            {
                "filename": str(assembly_path),
                "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<build>\d+)\.(?P<patch>\d+)",
                "serialize": ["{major}.{minor}.{build}.{patch}"],
                "search": '[assembly: AssemblyVersion("{current_version}")]',
                "replace": '[assembly: AssemblyVersion("{new_version}")]',
            },
            {
                "filename": str(csv_path),
                "parse": r"(?P<major>\d+);(?P<minor>\d+);(?P<patch>\d+);(?P<release>[0-9A-Za-z]+)?;(build\.(?P<build>.[0-9A-Za-z]+))?",
                "serialize": [
                    "{major};{minor};{patch};{release};build.{build}",
                    "{major};{minor};{patch};;build.{build}",
                ],
                "search": "1;{current_version}",
                "replace": "1;{new_version}",
            },
        ],
        "parts": {
            "release": {"values": ["beta", "rc", "final"], "optional_value": "final"},
            "build": {
                "first_value": 1000,
                "independent": True,
            },
        },
    }
    conf, version_config, current_version = get_config_data(overrides)
    major_version = current_version.bump("patch", version_config.order)

    ctx = get_context(conf)

    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.replace_version(current_version, major_version, ctx)

    assert full_vers_path.read_text() == "3.1.1-beta+build.1031"
    expected = '[assembly: AssemblyFileVersion("3.1.1-beta+build.1031")]\n[assembly: AssemblyVersion("3.1.1031.1")]'
    assert assembly_path.read_text() == expected
    assert csv_path.read_text() == "1;3;1;1;beta;build.1031"


def test_search_replace_to_avoid_updating_unconcerned_lines(tmp_path: Path):
    req_path = tmp_path / "requirements.txt"
    req_path.write_text("Django>=1.5.6,<1.6\nMyProject==1.5.6")

    changelog_path = tmp_path / "CHANGELOG.md"
    changelog_path.write_text(
        dedent(
            """
    # https://keepachangelog.com/en/1.0.0/

    ## [Unreleased]
    ### Added
    - Foobar

    ## [0.0.1] - 2014-05-31
    ### Added
    - This CHANGELOG file to hopefully serve as an evolving example of a
      standardized open source project CHANGELOG.
    """
        )
    )

    overrides = {
        "current_version": "1.5.6",
        "files": [
            {
                "filename": str(req_path),
                "search": "MyProject=={current_version}",
                "replace": "MyProject=={new_version}",
            },
            {
                "filename": str(changelog_path),
                "search": "## [Unreleased]",
                "replace": "## [Unreleased]\n\n## [{new_version}] - {utcnow:%Y-%m-%d}",
            },
        ],
    }

    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("minor", version_config.order)

    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.replace_version(current_version, new_version, get_context(conf))

    utc_today = datetime.now(timezone.utc).strftime("%Y-%m-%d")
    expected_chglog = dedent(
        f"""
    # https://keepachangelog.com/en/1.0.0/

    ## [Unreleased]

    ## [1.6.0] - {utc_today}
    ### Added
    - Foobar

    ## [0.0.1] - 2014-05-31
    ### Added
    - This CHANGELOG file to hopefully serve as an evolving example of a
      standardized open source project CHANGELOG.
    """
    )

    assert req_path.read_text() == "Django>=1.5.6,<1.6\nMyProject==1.6.0"
    assert changelog_path.read_text() == expected_chglog


def test_non_matching_search_does_not_modify_file(tmp_path: Path):
    changelog_path = tmp_path / "CHANGELOG.md"
    changelog_content = dedent(
        """
    # Unreleased

    * bullet point A

    # Release v'older' (2019-09-17)

    * bullet point B
    """
    )
    changelog_path.write_text(changelog_content)

    overrides = {
        "current_version": "1.0.3",
        "files": [
            {
                "filename": str(changelog_path),
                "search": "Not-yet-released",
                "replace": "Release v{new_version} ({now:%Y-%m-%d})",
            }
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("patch", version_config.order)
    configured_files = files.resolve_file_config(conf.files, version_config)
    with pytest.raises(exceptions.VersionNotFoundError, match="Did not find 'Not-yet-released' in file:"):
        files.modify_files(configured_files, current_version, new_version, get_context(conf))

    assert changelog_path.read_text() == changelog_content


def test_simple_replacement_in_utf8_file(tmp_path: Path):
    """Changing a file in UTF-8 should not change the non-ASCII characters."""
    # Arrange
    version_path = tmp_path / "VERSION"
    version_path.write_bytes("Kröt1.3.0".encode("utf-8"))

    overrides = {"current_version": "1.3.0", "files": [{"filename": str(version_path)}]}
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("patch", version_config.order)

    # Act
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.replace_version(current_version, new_version, get_context(conf))

    # Assert
    out = version_path.read_text()
    assert out == "Kröt1.3.1"


def test_multi_line_search_is_found(tmp_path: Path) -> None:
    """A multiline search string is found and replaced."""
    # Arrange
    alphabet_path = tmp_path / "the_alphabet.txt"
    alphabet_path.write_text("A\nB\nC\n")

    overrides = {
        "current_version": "9.8.7",
        "search": "A\nB\nC",
        "replace": "A\nB\nC\n{new_version}",
        "files": [{"filename": str(alphabet_path)}],
    }
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)
        new_version = current_version.bump("major", version_config.order)

    # Act
    for file_cfg in conf.files:
        cfg_file = files.ConfiguredFile(file_cfg, version_config)
        cfg_file.replace_version(current_version, new_version, get_context(conf))

    # Assert
    assert alphabet_path.read_text() == "A\nB\nC\n10.0.0\n"



from pathlib import Path

import pytest
from click import UsageError
from pytest import LogCaptureFixture, param

from bumpversion import config, exceptions
from bumpversion.utils import get_context
from bumpversion.version_part import VersionPart
from tests.conftest import get_config_data, inside_dir


@pytest.fixture(
    params=[
        None,
        ("0", "1", "2"),
        ("0", "3"),
    ]
)
def version_part_config(request):
    """Return a three-part and a two-part version part configuration."""
    if request.param is None:
        return config.VersionPartConfig()
    else:
        return config.VersionPartConfig(values=request.param)


def test_version_part_init(version_part_config):
    vp = VersionPart(version_part_config, version_part_config.first_value)
    assert vp.value is not None


def test_version_part_copy(version_part_config):
    vp = VersionPart(version_part_config, version_part_config.first_value)
    vc = vp.copy()
    assert vp.value == vc.value
    assert id(vp) != id(vc)


def test_version_part_bump(version_part_config):
    vp = VersionPart(version_part_config, version_part_config.first_value)
    vc = vp.bump()
    if version_part_config.values:
        assert vc.value == str(version_part_config.values[1])
    else:
        assert vc.value == "1"


def test_version_part_check_optional_false(version_part_config):
    assert not VersionPart(version_part_config, version_part_config.first_value).bump().is_optional


def test_version_part_check_optional_true(version_part_config):
    assert VersionPart(version_part_config, version_part_config.first_value).is_optional


def test_version_part_equality(version_part_config):
    assert VersionPart(version_part_config, version_part_config.first_value) == VersionPart(
        version_part_config, version_part_config.first_value
    )


def test_version_part_null(version_part_config):
    assert VersionPart(version_part_config, version_part_config.first_value).null() == VersionPart(
        version_part_config, version_part_config.first_value
    )


def test_bump_version_missing_part():
    overrides = {
        "current_version": "1.0.0",
    }
    conf, version_config, current_version = get_config_data(overrides)
    with pytest.raises(exceptions.InvalidVersionPartError, match="No part named 'bugfix'"):
        current_version.bump("bugfix", version_config.order)


def test_build_number_configuration():
    overrides = {
        "current_version": "2.1.6-5123",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)\-(?P<build>\d+)",
        "serialize": ["{major}.{minor}.{patch}-{build}"],
        "parts": {
            "build": {
                "independent": True,
            }
        },
    }
    conf, version_config, current_version = get_config_data(overrides)

    build_version_bump = current_version.bump("build", version_config.order)
    assert build_version_bump.values["build"].value == "5124"
    assert build_version_bump.values["build"].is_independent

    major_version_bump = build_version_bump.bump("major", version_config.order)
    assert major_version_bump.values["build"].value == "5124"
    assert build_version_bump.values["build"].is_independent
    assert major_version_bump.values["major"].value == "3"

    build_version_bump = major_version_bump.bump("build", version_config.order)
    assert build_version_bump.values["build"].value == "5125"
    assert build_version_bump.values["build"].is_independent


def test_serialize_with_distance_to_latest_tag():
    """Using ``distance_to_latest_tag`` in the serialization string outputs correctly."""
    from bumpversion.scm import Git, SCMInfo

    overrides = {
        "current_version": "19.6.0",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
        "serialize": ["{major}.{minor}.{patch}-pre{distance_to_latest_tag}"],
    }
    conf, version_config, current_version = get_config_data(overrides)
    conf.scm_info = SCMInfo(
        tool=Git, commit_sha="1234123412341234", distance_to_latest_tag=3, current_version="19.6.0", dirty=False
    )
    assert version_config.serialize(current_version, get_context(conf)) == "19.6.0-pre3"


def test_serialize_with_environment_var():
    import os

    os.environ["BUILD_NUMBER"] = "567"
    overrides = {
        "current_version": "2.3.4",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
        "serialize": ["{major}.{minor}.{patch}-pre{$BUILD_NUMBER}"],
    }
    conf, version_config, current_version = get_config_data(overrides)
    assert version_config.serialize(current_version, get_context(conf)) == "2.3.4-pre567"
    del os.environ["BUILD_NUMBER"]


def test_serialize_with_newlines():
    overrides = {
        "current_version": "MAJOR=31\nMINOR=0\nPATCH=3\n",
        "parse": r"MAJOR=(?P<major>\d+)\nMINOR=(?P<minor>\d+)\nPATCH=(?P<patch>\d+)\n",
        "serialize": ["MAJOR={major}\nMINOR={minor}\nPATCH={patch}\n"],
    }
    conf, version_config, current_version = get_config_data(overrides)
    new_version = current_version.bump("major", version_config.order)
    assert version_config.serialize(new_version, get_context(conf)) == "MAJOR=32\nMINOR=0\nPATCH=0\n"


@pytest.mark.parametrize(
    ["bump_type", "expected"],
    [
        param("patch", "0.9.1", id="patch"),
        param("minor", "0.10", id="minor"),
        param("major", "1", id="major"),
    ],
)
def test_serialize_three_part(bump_type: str, expected: str):
    overrides = {
        "current_version": "0.9",
        "parse": r"(?P<major>\d+)(\.(?P<minor>\d+)(\.(?P<patch>\d+))?)?",
        "serialize": ["{major}.{minor}.{patch}", "{major}.{minor}", "{major}"],
    }
    conf, version_config, current_version = get_config_data(overrides)

    new_version = current_version.bump(bump_type, version_config.order)
    assert version_config.serialize(new_version, get_context(conf)) == expected


@pytest.mark.parametrize(
    ["initial", "bump_type", "expected"],
    [
        param("1.5.dev", "release", "1.5", id="release"),
        param("1.5", "minor", "1.6.dev", id="minor"),
    ],
)
def test_bump_non_numeric_parts(initial: str, bump_type: str, expected: str):
    overrides = {
        "current_version": initial,
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<release>[a-z]+))?",
        "serialize": ["{major}.{minor}.{release}", "{major}.{minor}"],
        "parts": {
            "release": {
                "optional_value": "gamma",
                "values": ["dev", "gamma"],
            },
        },
    }
    conf, version_config, current_version = get_config_data(overrides)

    new_version = current_version.bump(bump_type, version_config.order)
    assert version_config.serialize(new_version, get_context(conf)) == expected


@pytest.mark.parametrize(
    ["initial", "bump_type", "expected"],
    [
        param("1.alpha", "release", "1.beta", id="alpha-to-beta-release"),
        param("1.beta", "release", "1", id="beta-to-release"),
    ],
)
def test_optional_value(initial: str, bump_type: str, expected: str):
    overrides = {
        "current_version": initial,
        "parse": r"(?P<num>\d+)(\.(?P<release>.*))?(\.)?",
        "serialize": ["{num}.{release}", "{num}"],
        "parts": {
            "release": {
                "optional_value": "gamma",
                "values": ["alpha", "beta", "gamma"],
            },
        },
    }
    conf, version_config, current_version = get_config_data(overrides)

    new_version = current_version.bump(bump_type, version_config.order)
    assert version_config.serialize(new_version, get_context(conf)) == expected


@pytest.mark.parametrize(
    ["initial", "bump_type", "expected"],
    [
        param("1.0a", "prerel", "1.0b", id="a-to-b-release"),
        param("1.0b", "prerelversion", "1.0b1", id="prerelease-version"),
        param("1.0b1", "prerelversion", "1.0b2", id="prerelease-version2"),
        param("1.0b2", "prerel", "1.0c", id="b2-to-c-release"),
        param("1.0c", "prerel", "1.0rc", id="c-to-rc-release"),
        param("1.0rc", "prerel", "1.0", id="rc-to-d-release"),
        param("1.0", "minor", "1.1dev", id="minor-release"),
        param("1.1dev", "prerel", "1.1a", id="dev-to-a-release"),
    ],
)
def test_python_pre_release_release_post_release(initial: str, bump_type: str, expected: str):
    # adapted from http://legacy.python.org/dev/peps/pep-0386/#the-new-versioning-algorithm
    overrides = {
        "current_version": initial,
        "parse": r"""^
            (?P<major>\d+)\.(?P<minor>\d+)   # minimum 'N.N'
            (?:
                (?P<prerel>[abc]|rc|dev)     # 'a' = alpha, 'b' = beta
                                             # 'c' or 'rc' = release candidate
                (?:
                    (?P<prerelversion>\d+(?:\.\d+)*)
                )?
            )?
            (?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?""",
        "serialize": [
            "{major}.{minor}{prerel}{prerelversion}",
            "{major}.{minor}{prerel}",
            "{major}.{minor}",
        ],
        "parts": {
            "prerel": {
                "optional_value": "d",
                "values": ["dev", "a", "b", "c", "rc", "d"],
            },
        },
    }
    conf, version_config, current_version = get_config_data(overrides)

    new_version = current_version.bump(bump_type, version_config.order)
    assert version_config.serialize(new_version, get_context(conf)) == expected


@pytest.mark.parametrize(
    ["initial", "bump_type", "expected"],
    [
        param("0.9.4", "major", "1.1.0", id="first-value-1"),
    ],
)
def test_part_first_value(initial: str, bump_type: str, expected: str):
    overrides = {
        "current_version": initial,
        "parts": {
            "minor": {
                "first_value": "1",
            },
        },
    }
    conf, version_config, current_version = get_config_data(overrides)

    new_version = current_version.bump(bump_type, version_config.order)
    assert version_config.serialize(new_version, get_context(conf)) == expected


def test_version_part_invalid_regex_exit(tmp_path: Path) -> None:
    """A version part with an invalid regex should raise an exception."""
    # Arrange
    overrides = {
        "current_version": "12",
        "parse": "*kittens*",
    }
    with inside_dir(tmp_path):
        with pytest.raises(UsageError):
            get_config_data(overrides)


def test_parse_doesnt_parse_current_version(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    """A warning should be output when the parse regex doesn't parse the version."""
    overrides = {"current_version": "12", "parse": "xxx"}
    with inside_dir(tmp_path):
        get_config_data(overrides)

    assert "Evaluating 'parse' option: 'xxx' does not parse current version '12'" in caplog.messages


def test_part_does_not_revert_to_zero_if_optional(tmp_path: Path) -> None:
    """A non-numeric part with the optional value should not revert to zero."""
    # From https://github.com/c4urself/bump2version/issues/248
    # Arrange
    overrides = {
        "current_version": "0.3.1",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)((?P<release>\D+)(?P<build>\d*))?",
        "serialize": [
            "{major}.{minor}.{patch}{release}{build}",
            "{major}.{minor}.{patch}{release}",
            "{major}.{minor}.{patch}",
        ],
        "parts": {
            "release": {
                "optional_value": "g",
                "first_value": "g",
                "values": [
                    "dev",
                    "a",
                    "b",
                    "g",
                ],
            },
        },
    }
    with inside_dir(tmp_path):
        conf, version_config, current_version = get_config_data(overrides)

    new_version = current_version.bump("build", version_config.order)
    assert version_config.serialize(new_version, get_context(conf)) == "0.3.1g1"



"""Module for managing Versions and their internal parts."""
import logging
import re
import string
from copy import copy
from typing import Any, Dict, List, MutableMapping, Optional

from click import UsageError

from bumpversion.config import VersionPartConfig
from bumpversion.exceptions import FormattingError, InvalidVersionPartError, MissingValueError
from bumpversion.functions import NumericFunction, PartFunction, ValuesFunction
from bumpversion.utils import key_val_string, labels_for_format

logger = logging.getLogger(__name__)


class VersionPart:
    """
    Represent part of a version number.

    Determines the PartFunction that rules how the part behaves when increased or reset
    based on the configuration given.
    """

    def __init__(self, config: VersionPartConfig, value: Optional[str] = None):
        self._value = value
        self.config = config
        self.func: Optional[PartFunction] = None
        if config.values:
            self.func = ValuesFunction(config.values, config.optional_value, config.first_value)
        else:
            self.func = NumericFunction(config.optional_value, config.first_value or "0")

    @property
    def value(self) -> str:
        """Return the value of the part."""
        return self._value or self.func.optional_value

    def copy(self) -> "VersionPart":
        """Return a copy of the part."""
        return VersionPart(self.config, self._value)

    def bump(self) -> "VersionPart":
        """Return a part with bumped value."""
        return VersionPart(self.config, self.func.bump(self.value))

    def null(self) -> "VersionPart":
        """Return a part with first value."""
        return VersionPart(self.config, self.func.first_value)

    @property
    def is_optional(self) -> bool:
        """Is the part optional?"""
        return self.value == self.func.optional_value

    @property
    def is_independent(self) -> bool:
        """Is the part independent of the other parts?"""
        return self.config.independent

    def __format__(self, format_spec: str) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"<bumpversion.VersionPart:{self.func.__class__.__name__}:{self.value}>"

    def __eq__(self, other: Any) -> bool:
        return self.value == other.value if isinstance(other, VersionPart) else False


class Version:
    """The specification of a version and its parts."""

    def __init__(self, values: Dict[str, VersionPart], original: Optional[str] = None):
        self.values = values
        self.original = original

    def __getitem__(self, key: str) -> VersionPart:
        return self.values[key]

    def __len__(self) -> int:
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def __repr__(self):
        return f"<bumpversion.Version:{key_val_string(self.values)}>"

    def __eq__(self, other: Any) -> bool:
        return (
            all(value == other.values[key] for key, value in self.values.items())
            if isinstance(other, Version)
            else False
        )

    def bump(self, part_name: str, order: List[str]) -> "Version":
        """Increase the value of the given part."""
        bumped = False

        new_values = {}

        for label in order:
            if label not in self.values:
                continue
            if label == part_name:
                new_values[label] = self.values[label].bump()
                bumped = True
            elif bumped and not self.values[label].is_independent:
                new_values[label] = self.values[label].null()
            else:
                new_values[label] = self.values[label].copy()

        if not bumped:
            raise InvalidVersionPartError(f"No part named {part_name!r}")

        return Version(new_values)


class VersionConfig:
    """
    Hold a complete representation of a version string.
    """

    def __init__(
        self,
        parse: str,
        serialize: List[str],
        search: str,
        replace: str,
        part_configs: Dict[str, VersionPartConfig] = None,
    ):
        try:
            self.parse_regex = re.compile(parse, re.VERBOSE)
        except re.error as e:
            raise UsageError(f"--parse '{parse}' is not a valid regex.") from e

        self.serialize_formats = serialize
        self.part_configs = part_configs or {}
        self.search = search
        self.replace = replace

    @property
    def order(self) -> List[str]:
        """
        Return the order of the labels in a serialization format.

        Currently, order depends on the first given serialization format.
        This seems like a good idea because this should be the most complete format.

        Returns:
            A list of version part labels in the order they should be rendered.
        """
        return labels_for_format(self.serialize_formats[0])

    def parse(self, version_string: Optional[str] = None) -> Optional[Version]:
        """
        Parse a version string into a Version object.

        Args:
            version_string: Version string to parse

        Returns:
            A Version object representing the string.
        """
        if not version_string:
            return None

        regexp_one_line = "".join([line.split("#")[0].strip() for line in self.parse_regex.pattern.splitlines()])

        logger.info(
            "Parsing version '%s' using regexp '%s'",
            version_string,
            regexp_one_line,
        )

        match = self.parse_regex.search(version_string)

        if not match:
            logger.warning(
                "Evaluating 'parse' option: '%s' does not parse current version '%s'",
                self.parse_regex.pattern,
                version_string,
            )
            return None

        _parsed = {
            key: VersionPart(self.part_configs[key], value)
            for key, value in match.groupdict().items()
            if key in self.part_configs
        }
        v = Version(_parsed, version_string)

        logger.info("Parsed the following values: %s", key_val_string(v.values))

        return v

    def _serialize(
        self, version: Version, serialize_format: str, context: MutableMapping, raise_if_incomplete: bool = False
    ) -> str:
        """
        Attempts to serialize a version with the given serialization format.

        Args:
            version: The version to serialize
            serialize_format: The serialization format to use, using Python's format string syntax
            context: The context to use when serializing the version
            raise_if_incomplete: Whether to raise an error if the version is incomplete

        Raises:
            FormattingError: if not serializable
            MissingValueError: if not all parts required in the format have values

        Returns:
            The serialized version as a string
        """
        values = copy(context)
        for k in version:
            values[k] = version[k]

        # TODO dump complete context on debug level

        try:
            # test whether all parts required in the format have values
            serialized = serialize_format.format(**values)

        except KeyError as e:
            missing_key = getattr(e, "message", e.args[0])
            raise MissingValueError(
                f"Did not find key {missing_key!r} in {repr(version)} when serializing version number"
            ) from e

        keys_needing_representation = set()

        keys = list(self.order)
        for i, k in enumerate(keys):
            v = values[k]

            if not isinstance(v, VersionPart):
                # values coming from environment variables don't need
                # representation
                continue

            if not v.is_optional:
                keys_needing_representation = set(keys[: i + 1])

        required_by_format = set(labels_for_format(serialize_format))

        # try whether all parsed keys are represented
        if raise_if_incomplete and not keys_needing_representation <= required_by_format:
            missing_keys = keys_needing_representation ^ required_by_format
            raise FormattingError(
                f"""Could not represent '{"', '".join(missing_keys)}' in format '{serialize_format}'"""
            )

        return serialized

    def _choose_serialize_format(self, version: Version, context: MutableMapping) -> str:
        chosen = None

        logger.debug("Available serialization formats: '%s'", "', '".join(self.serialize_formats))

        for serialize_format in self.serialize_formats:
            try:
                self._serialize(version, serialize_format, context, raise_if_incomplete=True)
                # Prefer shorter or first search expression.
                chosen_part_count = len(list(string.Formatter().parse(chosen))) if chosen else None
                serialize_part_count = len(list(string.Formatter().parse(serialize_format)))
                if not chosen or chosen_part_count > serialize_part_count:
                    chosen = serialize_format
                    logger.debug("Found '%s' to be a usable serialization format", chosen)
                else:
                    logger.debug("Found '%s' usable serialization format, but it's longer", serialize_format)
            except FormattingError:
                # If chosen, prefer shorter
                if not chosen:
                    chosen = serialize_format
            except MissingValueError as e:
                logger.info(e.message)
                raise e

        if not chosen:
            raise KeyError("Did not find suitable serialization format")

        logger.debug("Selected serialization format '%s'", chosen)

        return chosen

    def serialize(self, version: Version, context: MutableMapping) -> str:
        """
        Serialize a version to a string.

        Args:
            version: The version to serialize
            context: The context to use when serializing the version

        Returns:
            The serialized version as a string
        """
        serialized = self._serialize(version, self._choose_serialize_format(version, context), context)
        logger.debug("Serialized to '%s'", serialized)
        return serialized
